import os
import socket
import threading
from io import StringIO
from typing import Any
from unittest.mock import AsyncMock

import asyncssh
import paramiko
import pytest
from dulwich.contrib.test_paramiko_vendor import CLIENT_KEY, PASSWORD, USER, Server
from pytest_mock import MockerFixture
from pytest_test_utils.waiters import wait_until

from scmrepo.git.backend.dulwich.asyncssh_vendor import AsyncSSHVendor

# pylint: disable=redefined-outer-name


@pytest.fixture
def ssh_conn(request: pytest.FixtureRequest) -> dict[str, Any]:
    server = Server([])

    socket.setdefaulttimeout(10)
    request.addfinalizer(lambda: socket.setdefaulttimeout(None))

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    sock.listen(5)
    request.addfinalizer(sock.close)  # noqa: PT021
    port = sock.getsockname()[1]

    conn_info: dict[str, Any] = {"port": port, "server": server}

    def _run_server():
        try:
            conn, _ = sock.accept()
        except OSError:
            return False
        server.transport = transport = paramiko.Transport(conn)
        request.addfinalizer(transport.close)
        host_key = paramiko.RSAKey.from_private_key(StringIO(CLIENT_KEY))
        transport.add_server_key(host_key)
        transport.start_server(server=server)

    thread = threading.Thread(target=_run_server)
    thread.start()
    return conn_info


@pytest.fixture
def ssh_port(ssh_conn: dict[str, Any]) -> int:
    return ssh_conn["port"]


@pytest.fixture
def server(ssh_conn: dict[str, Any]) -> Server:
    return ssh_conn["server"]


def test_run_command_password(server: Server, ssh_port: int):
    vendor = AsyncSSHVendor()
    vendor.run_command(
        "127.0.0.1",
        "test_run_command_password",
        username=USER,
        port=ssh_port,
        password=PASSWORD,
    )

    assert b"test_run_command_password" in server.commands


def test_run_command_with_privkey(server: Server, ssh_port: int):
    key = asyncssh.import_private_key(CLIENT_KEY)

    vendor = AsyncSSHVendor()
    vendor.run_command(
        "127.0.0.1",
        "test_run_command_with_privkey",
        username=USER,
        port=ssh_port,
        key_filename=key,
    )

    assert b"test_run_command_with_privkey" in server.commands


@pytest.mark.parametrize("use_len", [True, False])
def test_run_command_data_transfer(server: Server, ssh_port: int, use_len: bool):
    vendor = AsyncSSHVendor()
    con = vendor.run_command(
        "127.0.0.1",
        "test_run_command_data_transfer",
        username=USER,
        port=ssh_port,
        password=PASSWORD,
    )

    assert b"test_run_command_data_transfer" in server.commands

    channel = server.transport.accept(5)  # type: ignore[attr-defined]
    channel.send(b"stdout\n")
    channel.send_stderr(b"stderr\n")
    channel.close()

    assert wait_until(con.can_read, timeout=1, pause=0.1)
    assert con.read(n=7 if use_len else None) == b"stdout\n"
    assert con.read_stderr(n=7 if use_len else None) == b"stderr\n"


def test_run_command_partial_transfer(ssh_port: int, mocker: MockerFixture):
    vendor = AsyncSSHVendor()
    con = vendor.run_command(
        "127.0.0.1",
        "test_run_command_data_transfer",
        username=USER,
        port=ssh_port,
        password=PASSWORD,
    )

    mock_stdout = mocker.patch.object(
        con.proc.stdout,
        "read",
        side_effect=[b"s", b"tdout", b"\n"],
        new_callable=AsyncMock,
    )
    assert con.read(n=7) == b"stdout\n"
    assert mock_stdout.call_count == 3

    mock_stderr = mocker.patch.object(
        con.stderr.stderr,
        "read",
        side_effect=[b"s", b"tderr", b"\n"],
        new_callable=AsyncMock,
    )
    assert con.read_stderr(n=7) == b"stderr\n"
    assert mock_stderr.call_count == 3


@pytest.mark.parametrize("algorithm", [b"ssh-rsa", b"rsa-sha2-256", b"rsa-sha2-512"])
def test_dulwich_github_compat(mocker: MockerFixture, algorithm: bytes):
    from asyncssh.misc import ProtocolError

    from scmrepo.git.backend.dulwich.asyncssh_vendor import _process_public_key_ok_gh

    key_data = b"foo"
    auth = mocker.Mock(
        _keypair=mocker.Mock(algorithm=algorithm, public_data=key_data),
    )
    packet = mocker.Mock()

    strings = iter((b"ed21556", key_data))
    packet.get_string = lambda: next(strings)
    with pytest.raises(ProtocolError):
        _process_public_key_ok_gh(auth, None, None, packet)

    strings = iter((b"ssh-rsa", key_data))
    packet.get_string = lambda: next(strings)
    _process_public_key_ok_gh(auth, None, None, packet)


@pytest.mark.skipif(os.name != "nt", reason="Windows only")
def test_git_bash_ssh_vendor(mocker):
    from dulwich.client import SubprocessSSHVendor

    from scmrepo.git.backend.dulwich import _get_ssh_vendor

    mocker.patch.dict(os.environ, {"MSYSTEM": "MINGW64"})
    assert isinstance(_get_ssh_vendor(), SubprocessSSHVendor)

    del os.environ["MSYSTEM"]
    assert isinstance(_get_ssh_vendor(), AsyncSSHVendor)


def test_unsupported_config_ssh_vendor():
    from dulwich.client import SubprocessSSHVendor

    from scmrepo.git.backend.dulwich import _get_ssh_vendor

    config = os.path.expanduser(os.path.join("~", ".ssh", "config"))
    os.makedirs(os.path.dirname(config), exist_ok=True)

    with open(config, "wb") as fobj:
        fobj.write(
            b"""
Host *
    IdentityFile ~/.ssh/id_rsa
"""
        )
    assert isinstance(_get_ssh_vendor(), AsyncSSHVendor)

    with open(config, "wb") as fobj:
        fobj.write(
            b"""
Host *
    UseKeychain yes
"""
        )
    assert isinstance(_get_ssh_vendor(), SubprocessSSHVendor)



"""asyncssh SSH vendor for Dulwich."""
import asyncio
import os
from collections.abc import Coroutine, Iterator, Sequence
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    cast,
)

from asyncssh import SSHClient
from dulwich.client import SSHVendor

from scmrepo.asyn import BaseAsyncObject, sync_wrapper
from scmrepo.exceptions import AuthError

if TYPE_CHECKING:
    from pathlib import Path

    from asyncssh.auth import KbdIntPrompts, KbdIntResponse
    from asyncssh.config import ConfigPaths, FilePath
    from asyncssh.connection import SSHClientConnection
    from asyncssh.misc import MaybeAwait
    from asyncssh.process import SSHClientProcess
    from asyncssh.public_key import KeyPairListArg, SSHKey
    from asyncssh.stream import SSHReader


async def _read_all(read: Callable[[int], Coroutine], n: Optional[int] = None) -> bytes:
    if n is None:
        return await read(-1)
    result = []
    while n > 0:
        data = await read(n)
        result.append(data)
        n -= len(data)
    return b"".join(result)


class _StderrWrapper:
    def __init__(self, stderr: "SSHReader", loop: asyncio.AbstractEventLoop) -> None:
        self.stderr = stderr
        self.loop = loop

    async def _readlines(self) -> list[bytes]:
        lines = []
        while True:
            line = await self.stderr.readline()
            if not line:
                break
            lines.append(line)
        return lines

    async def _read(self, n: Optional[int] = None) -> bytes:
        if self.stderr.at_eof():
            return b""
        return await _read_all(self.stderr.read, n)

    read = sync_wrapper(_read)
    readlines = sync_wrapper(_readlines)


class AsyncSSHWrapper(BaseAsyncObject):
    def __init__(self, conn: "SSHClientConnection", proc: "SSHClientProcess", **kwargs):
        super().__init__(**kwargs)
        self.conn: "SSHClientConnection" = conn
        self.proc: "SSHClientProcess" = proc
        self.stderr = _StderrWrapper(proc.stderr, self.loop)

    def can_read(self) -> bool:
        # pylint:disable=protected-access
        return self.proc.stdout._session._recv_buf_len > 0

    async def _read(self, n: Optional[int] = None) -> bytes:
        if self.proc.stdout.at_eof():
            return b""
        return await _read_all(self.proc.stdout.read, n)

    read = sync_wrapper(_read)

    def read_stderr(self, n: Optional[int] = None) -> bytes:
        return self.stderr.read(n)

    async def _write(self, data: bytes) -> None:
        self.proc.stdin.write(data)
        await self.proc.stdin.drain()

    write = sync_wrapper(_write)

    async def _close(self) -> None:
        self.conn.close()
        await self.conn.wait_closed()

    close = sync_wrapper(_close)


# NOTE: Github's SSH server does not strictly comply with the SSH protocol.
# When validating a public key using the rsa-sha2-256 or rsa-sha2-512
# signature algorithms, RFC4252 + RFC8332 state that the server should respond
# with the same algorithm in SSH_MSG_USERAUTH_PK_OK. Github's server always
# returns "ssh-rsa" rather than the correct sha2 algorithm name (likely for
# backwards compatibility with old SSH client reasons). This behavior causes
# asyncssh to fail with a key-mismatch error (since asyncssh expects the server
# to behave properly).
#
# See also:
#   https://www.ietf.org/rfc/rfc4252.txt
#   https://www.ietf.org/rfc/rfc8332.txt
def _process_public_key_ok_gh(self, _pkttype, _pktid, packet):
    from asyncssh.misc import ProtocolError

    algorithm = packet.get_string()
    key_data = packet.get_string()
    packet.check_end()

    # pylint: disable=protected-access
    if (
        (
            algorithm == b"ssh-rsa"
            and self._keypair.algorithm
            not in (
                b"ssh-rsa",
                b"rsa-sha2-256",
                b"rsa-sha2-512",
            )
        )
        or (algorithm not in (b"ssh-rsa", self._keypair.algorithm))
        or key_data != self._keypair.public_data
    ):
        raise ProtocolError("Key mismatch")

    self.create_task(self._send_signed_request())
    return True


class InteractiveSSHClient(SSHClient):
    _conn: Optional["SSHClientConnection"] = None
    _keys_to_try: Optional[list["FilePath"]] = None
    _passphrases: dict[str, str]

    def __init__(self, *args, **kwargs):
        super(*args, **kwargs)
        _passphrases: dict[str, str] = {}

    def connection_made(self, conn: "SSHClientConnection") -> None:
        self._conn = conn
        self._keys_to_try = None

    def connection_lost(self, exc: Optional[Exception]) -> None:
        self._conn = None

    async def public_key_auth_requested(  # noqa: C901, PLR0912
        self,
    ) -> Optional["KeyPairListArg"]:
        from asyncssh.public_key import (
            _DEFAULT_KEY_FILES,
            KeyImportError,
            SSHLocalKeyPair,
            read_private_key,
            read_public_key,
        )

        if os.environ.get("GIT_TERMINAL_PROMPT") == "0":
            return None

        assert self._conn is not None
        if self._keys_to_try is None:
            self._keys_to_try = []
            options = self._conn._options  # pylint: disable=protected-access
            config = options.config
            client_keys = cast(Sequence["FilePath"], config.get("IdentityFile", ()))
            if not client_keys:
                client_keys = [
                    os.path.expanduser(os.path.join("~", ".ssh", path))
                    for path, cond in _DEFAULT_KEY_FILES
                    if cond
                ]
            for key_to_load in client_keys:
                try:
                    read_private_key(key_to_load, passphrase=options.passphrase)
                except KeyImportError as exc:
                    if str(exc).startswith("Passphrase"):
                        self._keys_to_try.append(key_to_load)
                except OSError:
                    pass

        while self._keys_to_try:
            key_to_load = self._keys_to_try.pop()
            pubkey_to_load = str(key_to_load) + ".pub"
            try:
                key = await self._read_private_key_interactive(key_to_load)
            except KeyImportError:
                continue
            try:
                pubkey = read_public_key(pubkey_to_load)
            except (OSError, KeyImportError):
                pubkey = None
            return SSHLocalKeyPair(key, pubkey)
        return None

    async def _read_private_key_interactive(self, path: "FilePath") -> "SSHKey":
        from getpass import getpass

        from asyncssh.public_key import (
            KeyEncryptionError,
            KeyImportError,
            read_private_key,
        )

        path = str(path)
        passphrase = self._passphrases.get(path)
        if passphrase:
            return read_private_key(path, passphrase=passphrase)

        loop = asyncio.get_running_loop()
        for _ in range(3):
            passphrase = await loop.run_in_executor(
                None, getpass, f"Enter passphrase for key '{path}': "
            )
            if passphrase:
                try:
                    key = read_private_key(path, passphrase=passphrase)
                    self._passphrases[path] = passphrase
                    return key
                except (KeyImportError, KeyEncryptionError):
                    pass
        raise KeyImportError("Incorrect passphrase")

    def kbdint_auth_requested(self) -> "MaybeAwait[Optional[str]]":
        return ""

    async def kbdint_challenge_received(  # pylint: disable=invalid-overridden-method
        self,
        name: str,
        instructions: str,
        lang: str,
        prompts: "KbdIntPrompts",
    ) -> Optional["KbdIntResponse"]:
        from getpass import getpass

        if os.environ.get("GIT_TERMINAL_PROMPT") == "0":
            return None

        def _getpass(prompt: str) -> str:
            return getpass(prompt=prompt).rstrip()

        if instructions:
            pass
        loop = asyncio.get_running_loop()
        return [
            await loop.run_in_executor(
                None, _getpass, f"({name}) {prompt}" if name else prompt
            )
            for prompt, _ in prompts
        ]


class AsyncSSHVendor(BaseAsyncObject, SSHVendor):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    async def _run_command(
        self,
        host: str,
        command: list[str],
        username: Optional[str] = None,
        port: Optional[int] = None,
        password: Optional[str] = None,
        key_filename: Optional[str] = None,
        **kwargs: object,
    ) -> AsyncSSHWrapper:
        """Connect to an SSH server.

        Run a command remotely and return a file-like object for interaction
        with the remote command.

        Args:
          host: Host name
          command: Command to run (as argv array)
          username: Optional ame of user to log in as
          port: Optional SSH port to use
          password: Optional ssh password for login or private key
          key_filename: Optional path to private keyfile
        """
        import asyncssh
        from asyncssh.auth import MSG_USERAUTH_PK_OK, _ClientPublicKeyAuth

        # pylint: disable=protected-access
        _ClientPublicKeyAuth._packet_handlers[
            MSG_USERAUTH_PK_OK
        ] = _process_public_key_ok_gh

        try:
            conn = await asyncssh.connect(
                host,
                port=port if port is not None else (),
                username=username if username is not None else (),
                password=password if password is not None else (),
                client_keys=[key_filename] if key_filename else (),
                ignore_encrypted=not key_filename,
                known_hosts=None,
                encoding=None,
                client_factory=InteractiveSSHClient,
            )
            proc = await conn.create_process(command, encoding=None)
        except asyncssh.misc.PermissionDenied as exc:
            raise AuthError(f"{username}@{host}:{port or 22}") from exc
        return AsyncSSHWrapper(conn, proc)

    run_command = sync_wrapper(_run_command)


def get_unsupported_opts(config_paths: "ConfigPaths") -> Iterator[str]:
    from pathlib import Path, PurePath

    if config_paths:
        if isinstance(config_paths, (str, PurePath)):
            paths: Sequence["FilePath"] = [config_paths]
        else:
            paths = config_paths

        for path in paths:
            try:
                yield from _parse_unsupported(Path(path))
            except FileNotFoundError:
                continue


def _parse_unsupported(path: "Path") -> Iterator[str]:
    import locale
    import shlex

    from asyncssh.config import SSHClientConfig

    handlers = SSHClientConfig._handlers  # pylint: disable=protected-access
    with open(path, encoding=locale.getpreferredencoding()) as fobj:
        for line in fobj:
            line = line.strip()
            if not line or line[0] == "#":
                continue

            try:
                args = shlex.split(line)
            except ValueError:
                continue

            option = args.pop(0)
            if option.endswith("="):
                option = option[:-1]
            elif "=" in option:
                option, _ = option.split("=", 1)
            loption = option.lower()
            if loption not in handlers:
                yield loption



# pylint: disable=redefined-outer-name
import io

import pytest
from pytest_mock import MockerFixture
from pytest_test_utils import TempDirFactory, TmpDir

from scmrepo.git import Git
from scmrepo.git.lfs import LFSStorage, Pointer, smudge

FOO_OID = "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae"
FOO_POINTER = (
    f"version https://git-lfs.github.com/spec/v1\noid sha256:{FOO_OID}\nsize 3\n"
).encode()


@pytest.fixture
def storage(tmp_dir_factory: TempDirFactory) -> LFSStorage:
    storage_path = tmp_dir_factory.mktemp("lfs")
    return LFSStorage(storage_path)


@pytest.fixture
def lfs(tmp_dir: TmpDir, scm: Git) -> None:  # noqa: PT004
    tmp_dir.gen(".gitattributes", "*.lfs filter=lfs diff=lfs merge=lfs -text")
    scm.add([".gitattributes"])
    scm.commit("init lfs attributes")


@pytest.fixture
def lfs_objects(tmp_dir: TmpDir) -> TmpDir:
    objects = tmp_dir / ".git" / "lfs" / "objects"
    objects.mkdir(parents=True)
    return objects


def test_pointer_build(tmp_dir: TmpDir):
    tmp_dir.gen("foo", "foo")
    with open(tmp_dir / "foo", "rb") as fobj:
        pointer = Pointer.build(fobj)

    assert pointer.dump() == FOO_POINTER.decode("utf-8")


def test_pointer_load(tmp_dir: TmpDir):
    tmp_dir.gen("foo.lfs", FOO_POINTER)
    with open(tmp_dir / "foo.lfs", "rb") as fobj:
        pointer = Pointer.load(fobj)
    assert pointer.oid == FOO_OID
    assert pointer.size == 3


def test_smudge(tmp_dir: TmpDir, storage: LFSStorage, mocker: MockerFixture):
    tmp_dir.gen("foo.lfs", FOO_POINTER)
    with open(tmp_dir / "foo.lfs", "rb") as fobj:
        assert smudge(storage, fobj).read() == FOO_POINTER

    mocker.patch.object(storage, "open", return_value=io.BytesIO(b"foo"))
    with open(tmp_dir / "foo.lfs", "rb") as fobj:
        assert smudge(storage, fobj).read() == b"foo"


@pytest.mark.usefixtures("lfs")
def test_lfs(tmp_dir: TmpDir, scm: Git, lfs_objects: TmpDir):
    # NOTE: scmrepo does not currently support LFS clean (writes), this writes
    # the pointer to the git odb (simulating an actual LFS clean)
    tmp_dir.gen("foo.lfs", FOO_POINTER)
    scm.add(["foo.lfs"])
    scm.commit("add foo")
    lfs_objects.gen({FOO_OID[0:2]: {FOO_OID[2:4]: {FOO_OID: "foo"}}})

    fs = scm.get_fs("HEAD")
    with fs.open("foo.lfs", "rb", raw=True) as fobj:
        assert fobj.read() == FOO_POINTER
    with fs.open("foo.lfs", "rb", raw=False) as fobj:
        assert fobj.read() == b"foo"



import os
import shutil
from typing import Any, Optional

import pytest
from asyncssh import SFTPClient
from asyncssh.connection import SSHClientConnection
from dulwich.client import LocalGitClient
from git import Repo as GitPythonRepo
from pygit2 import GitError
from pygit2.remotes import Remote
from pytest_mock import MockerFixture
from pytest_test_utils import TempDirFactory, TmpDir
from pytest_test_utils.matchers import Matcher

from scmrepo.exceptions import InvalidRemote, MergeConflictError, RevError, SCMError
from scmrepo.git import Git
from scmrepo.git.objects import GitTag

from .conftest import backends

# pylint: disable=redefined-outer-name,unused-argument,protected-access


@pytest.fixture
def submodule_dir(tmp_dir: TmpDir, scm: Git):
    scm.commit("init")

    subrepo = GitPythonRepo.init()
    subrepo_path = "subrepo"
    subrepo.create_submodule(subrepo_path, subrepo_path, subrepo.git_dir)
    subrepo.close()

    return tmp_dir / subrepo_path


def test_git_init(tmp_dir: TmpDir, git_backend: str):
    Git.init(".", _backend=git_backend)
    assert (tmp_dir / ".git").is_dir()

    for backend in backends:
        Git(tmp_dir, backends=[backend])


def test_git_init_bare(tmp_dir: TmpDir, git_backend: str):
    Git.init(".", bare=True, _backend=git_backend)
    assert list(tmp_dir.iterdir())

    for backend in backends:
        Git(tmp_dir, backends=[backend])


def test_git_submodule(submodule_dir: TmpDir, git_backend: str):
    git = Git(backends=[git_backend])
    git.close()

    git = Git(submodule_dir, backends=[git_backend])
    git.close()


@pytest.mark.skip_git_backend("pygit2")
def test_commit(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"foo": "foo"})
    git.add(["foo"])
    git.commit("add")
    assert "foo" in scm.gitpython.git.ls_files()


@pytest.mark.skip_git_backend("pygit2")
def test_commit_in_root_repo_with_submodule(
    tmp_dir: TmpDir, scm: Git, git: Git, submodule_dir: TmpDir
):
    tmp_dir.gen("foo", "foo")
    git.add(["foo"])
    git.commit("add foo")
    assert "foo" in scm.gitpython.git.ls_files()


@pytest.mark.parametrize(
    "path, expected",
    [
        (os.path.join("path", "to", ".gitignore"), True),
        (os.path.join("path", "to", ".git", "internal", "file"), True),
        (os.path.join("some", "non-.git", "file"), False),
    ],
    ids=["gitignore_file", "git_internal_file", "non_git_file"],
)
def test_belongs_to_scm(scm: Git, git: Git, path: str, expected: str):
    assert git.belongs_to_scm(path) == expected


@pytest.mark.skip_git_backend("pygit2")
def test_is_tracked(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen(
        {
            "tracked": "tracked",
            "dir": {"data": "data", "subdir": {"subdata": "subdata"}},
        },
    )
    scm.add_commit(["tracked", "dir"], message="add dirs and files")
    tmp_dir.gen({"untracked": "untracked", "dir": {"untracked": "untracked"}})

    # sanity check
    assert (tmp_dir / "untracked").exists()
    assert (tmp_dir / "tracked").exists()
    assert (tmp_dir / "dir" / "untracked").exists()
    assert (tmp_dir / "dir" / "data").exists()
    assert (tmp_dir / "dir" / "subdir" / "subdata").exists()

    assert not git.is_tracked("untracked")
    assert not git.is_tracked(os.path.join("dir", "untracked"))

    assert git.is_tracked("tracked")
    assert git.is_tracked("dir")
    assert git.is_tracked(os.path.join("dir", "data"))
    assert git.is_tracked(os.path.join("dir", "subdir"))
    assert git.is_tracked(os.path.join("dir", "subdir", "subdata"))


@pytest.mark.skip_git_backend("pygit2")
def test_is_tracked_unicode(tmp_dir: TmpDir, scm: Git, git: Git):
    files = tmp_dir.gen("ṭṝḁḉḵḗḋ", "tracked")
    scm.add_commit(files, message="add unicode")
    tmp_dir.gen("ṳṋṭṝḁḉḵḗḋ", "untracked")

    assert git.is_tracked("ṭṝḁḉḵḗḋ")
    assert not git.is_tracked("ṳṋṭṝḁḉḵḗḋ")


@pytest.mark.skip_git_backend("pygit2")
def test_is_tracked_func(tmp_dir, scm, git):
    tmp_dir.gen({"foo": "foo", "тест": "проверка"})
    scm.add(["foo", "тест"])

    abs_foo = os.path.abspath("foo")
    assert git.is_tracked(abs_foo)
    assert git.is_tracked("foo")
    assert git.is_tracked("тест")

    git.commit("add")
    assert git.is_tracked(abs_foo)
    assert git.is_tracked("foo")

    scm.gitpython.repo.index.remove(["foo"], working_tree=True)
    assert not git.is_tracked(abs_foo)
    assert not git.is_tracked("foo")
    assert not git.is_tracked("not-existing-file")


@pytest.mark.skip_git_backend("pygit2")
def test_no_commits(tmp_dir: TmpDir, scm: Git, git: Git):
    assert git.no_commits

    tmp_dir.gen("foo", "foo")
    scm.add_commit(["foo"], message="foo")

    assert not git.no_commits


@pytest.mark.skip_git_backend("dulwich")
def test_branch_revs(tmp_dir: TmpDir, scm: Git, git: Git):
    def _gen(i: int):
        tmp_dir.gen({"file": f"{i}"})
        scm.add_commit("file", message=f"{i}")
        return scm.get_rev()

    base, *others = (_gen(i) for i in range(5))
    branch_revs = list(git.branch_revs("master", base))[::-1]
    assert branch_revs == others


def test_set_ref(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen("file", "0")
    scm.add_commit("file", message="init")
    init_rev = scm.get_rev()

    tmp_dir.gen({"file": "1"})
    scm.add_commit("file", message="commit")
    commit_rev = scm.get_rev()

    git.set_ref("refs/foo/bar", init_rev)
    assert init_rev == (tmp_dir / ".git" / "refs" / "foo" / "bar").read_text().strip()

    with pytest.raises(SCMError):
        git.set_ref("refs/foo/bar", commit_rev, old_ref=commit_rev)
    git.set_ref("refs/foo/bar", commit_rev, old_ref=init_rev)
    assert commit_rev == (tmp_dir / ".git" / "refs" / "foo" / "bar").read_text().strip()

    git.set_ref("refs/foo/baz", "refs/heads/master", symbolic=True)
    assert (
        tmp_dir / ".git" / "refs" / "foo" / "baz"
    ).read_text().strip() == "ref: refs/heads/master"


def test_get_ref(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    init_rev = scm.get_rev()
    tmp_dir.gen(
        {
            os.path.join(".git", "refs", "foo", "bar"): init_rev,
            os.path.join(".git", "refs", "foo", "baz"): "ref: refs/heads/master",
        }
    )
    scm.tag("annotated", annotated=True, message="Annotated Tag")

    assert init_rev == git.get_ref("refs/foo/bar")
    assert init_rev == git.get_ref("refs/foo/baz")
    assert init_rev == git.get_ref("refs/tags/annotated", follow=True)
    assert init_rev == git.get_ref("refs/tags/annotated", follow=False)
    assert git.get_ref("refs/foo/baz", follow=False) == "refs/heads/master"
    assert git.get_ref("refs/foo/qux") is None


def test_remove_ref(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    init_rev = scm.get_rev()

    # remove nonexistent ref should silently pass when old_ref is None)
    git.remove_ref("refs/foo/bar", old_ref=None)

    tmp_dir.gen(os.path.join(".git", "refs", "foo", "bar"), init_rev)
    tmp_dir.gen({"file": "1"})
    scm.add_commit("file", message="commit")
    commit_rev = scm.get_rev()

    with pytest.raises(SCMError):
        git.remove_ref("refs/foo/bar", old_ref=commit_rev)
    git.remove_ref("refs/foo/bar", old_ref=init_rev)
    assert not (tmp_dir / ".git" / "refs" / "foo" / "bar").exists()


@pytest.mark.skip_git_backend("dulwich")
def test_refs_containing(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    init_rev = scm.get_rev()
    tmp_dir.gen(
        {
            os.path.join(".git", "refs", "foo", "bar"): init_rev,
            os.path.join(".git", "refs", "foo", "baz"): init_rev,
        }
    )

    expected = {"refs/foo/bar", "refs/foo/baz", "refs/heads/master"}
    assert expected == set(git.get_refs_containing(init_rev))


@pytest.mark.skip_git_backend("pygit2", "gitpython")
@pytest.mark.parametrize("use_url", [True, False])
def test_push_refspecs(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
    remote_git_dir: TmpDir,
    use_url: str,
):
    from scmrepo.git.backend.dulwich import SyncStatus

    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    init_rev = scm.get_rev()
    scm.add_commit("file", message="bar")
    bar_rev = scm.get_rev()
    scm.checkout(init_rev)
    scm.add_commit("file", message="baz")
    baz_rev = scm.get_rev()
    tmp_dir.gen(
        {
            os.path.join(".git", "refs", "foo", "bar"): bar_rev,
            os.path.join(".git", "refs", "foo", "baz"): baz_rev,
        }
    )

    url = f"file://{remote_git_dir.resolve().as_posix()}"
    remote_scm = Git(remote_git_dir)
    scm.gitpython.repo.create_remote("origin", url)

    with pytest.raises(SCMError):
        git.push_refspecs("bad-remote", "refs/foo/bar:refs/foo/bar")

    remote = url if use_url else "origin"
    assert git.push_refspecs(remote, "refs/foo/bar:refs/foo/bar") == {
        "refs/foo/bar": SyncStatus.SUCCESS
    }
    assert bar_rev == remote_scm.get_ref("refs/foo/bar")

    remote_scm.checkout("refs/foo/bar")
    assert bar_rev == remote_scm.get_rev()
    assert (remote_git_dir / "file").read_text() == "0"

    assert git.push_refspecs(
        remote, ["refs/foo/bar:refs/foo/bar", "refs/foo/baz:refs/foo/baz"]
    ) == {
        "refs/foo/bar": SyncStatus.UP_TO_DATE,
        "refs/foo/baz": SyncStatus.SUCCESS,
    }
    assert baz_rev == remote_scm.get_ref("refs/foo/baz")

    assert git.push_refspecs(remote, ["refs/foo/bar:refs/foo/baz"]) == {
        "refs/foo/baz": SyncStatus.DIVERGED
    }
    assert baz_rev == remote_scm.get_ref("refs/foo/baz")

    assert git.push_refspecs(remote, ":refs/foo/baz") == {
        "refs/foo/baz": SyncStatus.SUCCESS
    }
    assert remote_scm.get_ref("refs/foo/baz") is None


@pytest.mark.skip_git_backend("gitpython")
@pytest.mark.parametrize("use_url", [True, False])
def test_fetch_refspecs(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
    remote_git_dir: TmpDir,
    use_url: bool,
    mocker: MockerFixture,
):
    from scmrepo.git.backend.dulwich import SyncStatus

    url = f"file://{remote_git_dir.resolve().as_posix()}"
    scm.gitpython.repo.create_remote("origin", url)
    remote_scm = Git(remote_git_dir)
    remote_git_dir.gen("file", "0")

    remote_scm.add_commit("file", message="init")
    init_rev = remote_scm.get_rev()
    remote_scm.add_commit("file", message="bar")
    bar_rev = remote_scm.get_rev()
    remote_scm.checkout(init_rev)
    remote_scm.add_commit("file", message="baz")
    baz_rev = remote_scm.get_rev()
    remote_git_dir.gen(
        {
            os.path.join(".git", "refs", "foo", "bar"): bar_rev,
            os.path.join(".git", "refs", "foo", "baz"): baz_rev,
        }
    )

    with pytest.raises(SCMError):
        git.fetch_refspecs("bad-remote", "refs/foo/bar:refs/foo/bar")

    remote = url if use_url else "origin"
    assert git.fetch_refspecs(remote, "refs/foo/bar:refs/foo/bar") == {
        "refs/foo/bar": SyncStatus.SUCCESS
    }
    assert bar_rev == scm.get_ref("refs/foo/bar")

    assert git.fetch_refspecs(
        remote, ["refs/foo/bar:refs/foo/bar", "refs/foo/baz:refs/foo/baz"]
    ) == {
        "refs/foo/bar": SyncStatus.UP_TO_DATE,
        "refs/foo/baz": SyncStatus.SUCCESS,
    }
    assert baz_rev == scm.get_ref("refs/foo/baz")

    assert git.fetch_refspecs(remote, ["refs/foo/bar:refs/foo/baz"]) == {
        "refs/foo/baz": SyncStatus.DIVERGED
    }
    assert baz_rev == scm.get_ref("refs/foo/baz")

    mocker.patch.object(LocalGitClient, "fetch", side_effect=KeyError)
    mocker.patch.object(Remote, "fetch", side_effect=GitError)
    with pytest.raises(SCMError):
        git.fetch_refspecs(remote, "refs/foo/bar:refs/foo/bar")

    assert len(scm.pygit2.repo.remotes) == 1


@pytest.mark.skip_git_backend("pygit2", "gitpython")
@pytest.mark.parametrize("use_url", [True, False])
def test_iter_remote_refs(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
    remote_git_dir: TmpDir,
    tmp_dir_factory: TempDirFactory,
    use_url: bool,
):
    url = f"file://{remote_git_dir.resolve().as_posix()}"

    scm.gitpython.repo.create_remote("origin", url)
    remote_scm = Git(remote_git_dir)
    remote_git_dir.gen("file", "0")
    remote_scm.add_commit("file", message="init")
    remote_scm.branch("new-branch")
    remote_scm.add_commit("file", message="bar")
    remote_scm.add_commit("file", message="baz")
    remote_scm.tag("a-tag")

    with pytest.raises(InvalidRemote):
        set(git.iter_remote_refs("bad-remote"))

    tmp_directory = tmp_dir_factory.mktemp("not_a_git_repo")
    remote = f"file://{tmp_directory.as_posix()}"
    with pytest.raises(InvalidRemote):
        set(git.iter_remote_refs(remote))

    remote = url if use_url else "origin"
    assert {
        "refs/heads/master",
        "HEAD",
        "refs/heads/new-branch",
        "refs/tags/a-tag",
    } == set(git.iter_remote_refs(remote))


@pytest.mark.skip_git_backend("dulwich", "pygit2")
def test_list_all_commits(tmp_dir: TmpDir, scm: Git, git: Git, matcher: type[Matcher]):
    def _gen(s):
        tmp_dir.gen(s, s)
        scm.add_commit(s, message=s)
        return scm.get_rev()

    rev_a = _gen("a")
    rev_b = _gen("b")
    scm.tag("tag")
    rev_c = _gen("c")
    scm.gitpython.git.reset(rev_a, hard=True)
    scm.set_ref("refs/foo/bar", rev_c)

    assert git.list_all_commits() == matcher.unordered(rev_a, rev_b)


@pytest.mark.skip_git_backend("pygit2")
def test_ignore_remove_empty(tmp_dir: TmpDir, scm: Git, git: Git):
    test_entries = [
        {"entry": "/foo1", "path": f"{tmp_dir}/foo1"},
        {"entry": "/foo2", "path": f"{tmp_dir}/foo2"},
    ]

    path_to_gitignore = tmp_dir / ".gitignore"

    with open(path_to_gitignore, "a", encoding="utf-8") as f:
        for entry in test_entries:
            f.write(entry["entry"] + "\n")

    assert path_to_gitignore.exists()

    git.ignore_remove(test_entries[0]["path"])
    assert path_to_gitignore.exists()

    git.ignore_remove(test_entries[1]["path"])
    assert not path_to_gitignore.exists()


@pytest.mark.skip_git_backend("pygit2")
@pytest.mark.skipif(os.name == "nt", reason="Git hooks not supported on Windows")
@pytest.mark.parametrize("hook", ["pre-commit", "commit-msg"])
def test_commit_no_verify(tmp_dir: TmpDir, scm: Git, git: Git, hook: str):
    import stat

    hook_file = os.path.join(".git", "hooks", hook)
    tmp_dir.gen(hook_file, "#!/usr/bin/env python\nimport sys\nsys.exit(1)")
    os.chmod(hook_file, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC)

    tmp_dir.gen("foo", "foo")
    git.add(["foo"])
    with pytest.raises(SCMError):
        git.commit("commit foo")
    git.commit("commit foo", no_verify=True)


@pytest.mark.skip_git_backend("dulwich")
@pytest.mark.parametrize("squash", [True, False])
def test_merge(tmp_dir: TmpDir, scm: Git, git: Git, squash: bool):
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="init")
    init_rev = scm.get_rev()

    scm.checkout("branch", create_new=True)
    tmp_dir.gen("foo", "bar")
    scm.add_commit("foo", message="bar")
    branch = scm.resolve_rev("branch")

    scm.checkout("master")

    tmp_dir.gen("foo", "baz")
    scm.add_commit("foo", message="baz")
    with pytest.raises(MergeConflictError):
        git.merge(branch, commit=not squash, squash=squash, msg="merge")

    scm.gitpython.git.reset(init_rev, hard=True)
    merge_rev = git.merge(branch, commit=not squash, squash=squash, msg="merge")
    assert (tmp_dir / "foo").read_text() == "bar"
    if squash:
        assert merge_rev is None
        assert scm.get_rev() == init_rev
    else:
        assert scm.get_rev() == merge_rev


@pytest.mark.skip_git_backend("dulwich")
def test_checkout_index(tmp_dir: TmpDir, scm: Git, git: Git):
    files = tmp_dir.gen({"foo": "foo", "bar": "bar", "dir": {"baz": "baz"}})
    scm.add_commit(files, message="init")
    tmp_dir.gen({"foo": "baz", "dir": {"baz": "foo"}})

    with (tmp_dir / "dir").chdir():
        git.checkout_index([os.path.join("..", "foo"), "baz"], force=True)
    assert (tmp_dir / "foo").read_text() == "foo"
    assert (tmp_dir / "dir" / "baz").read_text() == "baz"

    tmp_dir.gen({"foo": "baz", "bar": "baz", "dir": {"baz": "foo"}})
    git.checkout_index(force=True)
    assert (tmp_dir / "foo").read_text() == "foo"
    assert (tmp_dir / "bar").read_text() == "bar"
    assert (tmp_dir / "dir" / "baz").read_text() == "baz"


@pytest.mark.skip_git_backend("dulwich")
@pytest.mark.parametrize("strategy, expected", [("ours", "baz"), ("theirs", "bar")])
def test_checkout_index_conflicts(
    tmp_dir: TmpDir, scm: Git, git: Git, strategy: str, expected: str
):
    tmp_dir.gen({"file": "foo"})
    scm.add_commit("file", message="init")

    scm.checkout("branch", create_new=True)
    tmp_dir.gen({"file": "bar"})
    scm.add_commit("file", message="bar")
    rev_bar = scm.get_rev()

    scm.checkout("master")
    tmp_dir.gen({"file": "baz"})
    scm.add_commit("file", message="baz")

    with pytest.raises(MergeConflictError):
        git.merge(rev_bar, commit=False, squash=True)

    git.checkout_index(
        ours=strategy == "ours",
        theirs=strategy == "theirs",
    )
    assert (tmp_dir / "file").read_text() == expected


@pytest.mark.skip_git_backend("dulwich")
def test_resolve_rev(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
    remote_git_dir: TmpDir,
):
    url = f"file://{remote_git_dir.resolve().as_posix()}"
    scm.gitpython.repo.create_remote("origin", url)
    scm.gitpython.repo.create_remote("upstream", url)

    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    init_rev = scm.get_rev()

    tmp_dir.gen({"file": "1"})
    scm.add_commit("file", message="1")
    rev = scm.get_rev()

    scm.checkout("branch", create_new=True)
    tmp_dir.gen(
        {
            os.path.join(".git", "refs", "foo"): rev,
            os.path.join(".git", "refs", "remotes", "origin", "bar"): rev,
            os.path.join(".git", "refs", "remotes", "origin", "baz"): rev,
            os.path.join(".git", "refs", "remotes", "upstream", "baz"): init_rev,
        }
    )

    assert git.resolve_rev(rev) == rev
    assert git.resolve_rev(rev[:7]) == rev
    assert git.resolve_rev("HEAD") == rev
    assert git.resolve_rev("branch") == rev
    assert git.resolve_rev("refs/foo") == rev
    assert git.resolve_rev("bar") == rev
    assert git.resolve_rev("origin/baz") == rev
    assert git.resolve_rev("HEAD~1") == init_rev

    with pytest.raises(RevError):
        git.resolve_rev("qux")

    with pytest.raises(RevError):
        git.resolve_rev("baz")

    with pytest.raises(RevError):
        git.resolve_rev("HEAD~3")


@pytest.mark.skip_git_backend("dulwich")
def test_checkout(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"foo": "foo"})
    scm.add_commit("foo", message="foo")
    foo_rev = scm.get_rev()

    tmp_dir.gen("foo", "bar")
    scm.add_commit("foo", message="bar")
    bar_rev = scm.get_rev()

    git.checkout("branch", create_new=True)
    assert git.get_ref("HEAD", follow=False) == "refs/heads/branch"
    assert (tmp_dir / "foo").read_text() == "bar"

    git.checkout("master", detach=True)
    assert git.get_ref("HEAD", follow=False) == bar_rev

    git.checkout("master")
    assert git.get_ref("HEAD", follow=False) == "refs/heads/master"

    git.checkout(foo_rev[:7])
    assert git.get_ref("HEAD", follow=False) == foo_rev
    assert (tmp_dir / "foo").read_text() == "foo"


@pytest.mark.skip_git_backend("dulwich")
def test_reset(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"foo": "foo", "dir": {"baz": "baz"}})
    scm.add_commit(["foo", "dir"], message="init")

    tmp_dir.gen({"foo": "bar", "dir": {"baz": "bar"}})
    scm.add(["foo", os.path.join("dir", "baz")])
    git.reset()
    assert (tmp_dir / "foo").read_text() == "bar"
    assert (tmp_dir / "dir" / "baz").read_text() == "bar"
    staged, unstaged, _ = scm.status()
    assert len(staged) == 0
    assert set(unstaged) == {"foo", "dir/baz"}

    scm.add(["foo", os.path.join("dir", "baz")])
    git.reset(hard=True)
    assert (tmp_dir / "foo").read_text() == "foo"
    assert (tmp_dir / "dir" / "baz").read_text() == "baz"
    staged, unstaged, _ = scm.status()
    assert len(staged) == 0
    assert len(unstaged) == 0

    tmp_dir.gen({"foo": "bar", "bar": "bar", "dir": {"baz": "bar"}})
    scm.add(["foo", "bar", os.path.join("dir", "baz")])
    with (tmp_dir / "dir").chdir():
        git.reset(paths=[os.path.join("..", "foo"), os.path.join("baz")])
    assert (tmp_dir / "foo").read_text() == "bar"
    assert (tmp_dir / "bar").read_text() == "bar"
    assert (tmp_dir / "dir" / "baz").read_text() == "bar"
    staged, unstaged, _ = scm.status()
    assert len(staged) == 1
    assert len(unstaged) == 2


@pytest.mark.skip_git_backend("pygit2")
def test_add(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"foo": "foo", "bar": "bar", "dir": {"baz": "baz"}})
    git.add(["foo", "dir"])
    staged, unstaged, untracked = scm.status()
    assert set(staged["add"]) == {"foo", "dir/baz"}
    assert len(unstaged) == 0
    assert len(untracked) == 1

    scm.commit("commit")
    tmp_dir.gen({"foo": "bar", "dir": {"baz": "bar"}})
    git.add([], update=True)
    staged, unstaged, _ = scm.status()
    assert set(staged["modify"]) == {"foo", "dir/baz"}
    assert len(unstaged) == 0
    assert len(untracked) == 1

    scm.reset()
    git.add(["dir"], update=True)
    staged, unstaged, _ = scm.status()
    assert set(staged["modify"]) == {"dir/baz"}
    assert len(unstaged) == 1
    assert len(untracked) == 1


@pytest.mark.skip_git_backend("pygit2")
def test_add_force(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"foo": "foo", "bar": "bar", "dir": {"baz": "baz"}})
    tmp_dir.gen({".gitignore": "foo\ndir/"})

    git.add(["foo", "bar", "dir"])
    staged, _unstaged, untracked = scm.status()
    assert set(staged["add"]) == {"bar"}
    assert set(untracked) == {".gitignore"}

    git.add(["foo", "bar", "dir"], force=True)
    staged, _unstaged, untracked = scm.status()
    assert set(staged["add"]) == {"foo", "bar", "dir/baz"}
    assert set(untracked) == {".gitignore"}


@pytest.mark.skip_git_backend("dulwich", "gitpython")
def test_checkout_subdir(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="init")
    rev = scm.get_rev()

    tmp_dir.gen({"dir": {"bar": "bar"}})
    scm.add_commit("dir", message="dir")

    with (tmp_dir / "dir").chdir():
        git.checkout(rev)
        assert not (tmp_dir / "dir" / "bar").exists()


@pytest.mark.skip_git_backend("pygit2", "gitpython")
def test_describe(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"foo": "foo"})
    scm.add_commit("foo", message="foo")
    rev_foo = scm.get_rev()

    tmp_dir.gen({"foo": "bar"})
    scm.add_commit("foo", message="bar")
    rev_bar = scm.get_rev()

    assert git.describe([rev_foo], "refs/heads") == {rev_foo: None}

    scm.checkout("branch", create_new=True)
    assert git.describe([rev_bar], "refs/heads") == {rev_bar: "refs/heads/branch"}

    tmp_dir.gen({"foo": "foobar"})
    scm.add_commit("foo", message="foobar")
    rev_foobar = scm.get_rev()

    scm.checkout("master")
    assert git.describe([rev_bar, rev_foobar], "refs/heads") == {
        rev_bar: "refs/heads/master",
        rev_foobar: "refs/heads/branch",
    }

    scm.tag("tag")
    assert git.describe([rev_bar, "nonexist"]) == {
        rev_bar: "refs/tags/tag",
        "nonexist": None,
    }


def test_ignore(tmp_dir: TmpDir, scm: Git, git: Git):
    file = os.fspath(tmp_dir / "foo")

    git.ignore(file)
    assert (tmp_dir / ".gitignore").cat() == "/foo\n"

    git._reset()
    git.ignore(file)
    assert (tmp_dir / ".gitignore").cat() == "/foo\n"

    git._reset()
    git.ignore_remove(file)
    assert not (tmp_dir / ".gitignore").exists()


def test_ignored(tmp_dir: TmpDir, scm: Git, git: Git, git_backend: str):
    tmp_dir.gen({"dir1": {"file1.jpg": "cont", "file2.txt": "cont"}})
    tmp_dir.gen({".gitignore": "dir1/*.jpg"})

    git._reset()

    assert git.is_ignored(tmp_dir / "dir1" / "file1.jpg")
    assert not git.is_ignored(tmp_dir / "dir1" / "file2.txt")


@pytest.mark.skip_git_backend("pygit2", "gitpython")
def test_ignored_dir_unignored_subdirs(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({".gitignore": "data/**\n!data/**/\n!data/**/*.csv"})
    scm.add([".gitignore"])
    tmp_dir.gen(
        {
            os.path.join("data", "raw", "tracked.csv"): "cont",
            os.path.join("data", "raw", "not_tracked.json"): "cont",
        }
    )

    git._reset()

    assert not git.is_ignored(tmp_dir / "data" / "raw" / "tracked.csv")
    assert git.is_ignored(tmp_dir / "data" / "raw" / "not_tracked.json")
    assert not git.is_ignored(tmp_dir / "data" / "raw" / "non_existent.csv")
    assert git.is_ignored(tmp_dir / "data" / "raw" / "non_existent.json")
    assert not git.is_ignored(tmp_dir / "data" / "non_existent.csv")
    assert git.is_ignored(tmp_dir / "data" / "non_existent.json")

    assert not git.is_ignored(f"data{os.sep}")
    # git check-ignore would now mark "data/raw" as ignored
    # after detecting it's a directory in the file system;
    # instead, we rely on the trailing separator to determine if handling a
    # a directory - for consistency between existent and non-existent paths
    assert git.is_ignored(os.path.join("data", "raw"))
    assert not git.is_ignored(os.path.join("data", f"raw{os.sep}"))

    assert git.is_ignored(os.path.join("data", "non_existent"))
    assert not git.is_ignored(os.path.join("data", f"non_existent{os.sep}"))


def test_get_gitignore(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"file1": "contents", "dir": {}})

    data_dir = os.fspath(tmp_dir / "file1")
    entry, gitignore = git._get_gitignore(data_dir)
    assert entry == "/file1"
    assert gitignore == os.fspath(tmp_dir / ".gitignore")

    data_dir = os.fspath(tmp_dir / "dir")
    entry, gitignore = git._get_gitignore(data_dir)

    assert entry == "/dir"
    assert gitignore == os.fspath(tmp_dir / ".gitignore")


def test_get_gitignore_symlink(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"dir": {"subdir": {"data": "contents"}}})
    link = tmp_dir / "link"
    link.symlink_to(tmp_dir / "dir" / "subdir" / "data")

    entry, gitignore = git._get_gitignore(os.fspath(link))
    assert entry == "/link"
    assert gitignore == os.fspath(tmp_dir / ".gitignore")


def test_get_gitignore_subdir(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"dir1": {"file1": "cont", "dir2": {}}})

    data_dir = os.fspath(tmp_dir / "dir1" / "file1")
    entry, gitignore = git._get_gitignore(data_dir)
    assert entry == "/file1"
    assert gitignore == os.fspath(tmp_dir / "dir1" / ".gitignore")

    data_dir = os.fspath(tmp_dir / "dir1" / "dir2")
    entry, gitignore = git._get_gitignore(data_dir)
    assert entry == "/dir2"
    assert gitignore == os.fspath(tmp_dir / "dir1" / ".gitignore")


def test_gitignore_should_append_newline_to_gitignore(
    tmp_dir: TmpDir, scm: Git, git: Git
):
    tmp_dir.gen({"foo": "foo", "bar": "bar"})

    bar_path = os.fspath(tmp_dir / "bar")
    gitignore = tmp_dir / ".gitignore"

    gitignore.write_text("/foo")
    assert not gitignore.read_text().endswith("\n")

    git.ignore(bar_path)
    contents = gitignore.read_text()
    assert gitignore.read_text().endswith("\n")

    assert contents.splitlines() == ["/foo", "/bar"]


@pytest.mark.skip_git_backend("dulwich")
def test_git_detach_head(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen({"file": "0"})
    scm.add_commit("file", message="init")
    init_rev = scm.get_rev()

    with git.detach_head() as rev:
        assert init_rev == rev
        assert init_rev == (tmp_dir / ".git" / "HEAD").read_text().strip()
    assert (tmp_dir / ".git" / "HEAD").read_text().strip() == "ref: refs/heads/master"


@pytest.mark.skip_git_backend("pygit2", "gitpython")
@pytest.mark.slow
@pytest.mark.asyncio
async def test_git_ssh(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
    sftp: SFTPClient,
    ssh_connection: SSHClientConnection,
    ssh_conn_info: dict[str, Any],
):
    host, port = ssh_conn_info["host"], ssh_conn_info["port"]
    key_filename = ssh_conn_info["client_keys"]
    user = ssh_conn_info["username"]
    url = f"ssh://{user}@{host}:{port}/~/test-repo.git"

    result = await ssh_connection.run("git init --bare test-repo.git")
    assert result.returncode == 0

    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="init")
    rev = scm.get_rev()

    git.push_refspecs(
        url,
        "refs/heads/master:refs/heads/master",
        force=True,
        key_filename=key_filename,
    )

    async with sftp.open("test-repo.git/refs/heads/master") as fobj:
        assert (await fobj.read(-1, 0)).strip() == rev

    shutil.rmtree(tmp_dir / ".git")
    (tmp_dir / "foo").unlink()
    scm = Git.init(str(tmp_dir))

    git.fetch_refspecs(
        url,
        ["refs/heads/master"],
        force=True,
        key_filename=key_filename,
    )
    assert git.get_ref("refs/heads/master") == rev
    scm.checkout("master")
    assert (tmp_dir / "foo").read_text() == "foo"


@pytest.mark.parametrize("scheme", ["", "file://"])
@pytest.mark.parametrize("shallow_branch", [None, "master"])
@pytest.mark.parametrize("bare", [True, False])
def test_clone(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
    tmp_dir_factory: TempDirFactory,
    scheme: str,
    shallow_branch: Optional[str],
    bare: bool,
):
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="init")
    rev = scm.get_rev()

    target_dir = tmp_dir_factory.mktemp("git-clone")
    git.clone(
        f"{scheme}{tmp_dir}",
        str(target_dir),
        shallow_branch=shallow_branch,
        bare=bare,
    )
    target = Git(str(target_dir))
    assert target.get_rev() == rev
    if bare:
        assert not (target_dir / "foo").exists()
    else:
        assert (target_dir / "foo").read_text() == "foo"
        assert (target_dir / "foo").read_text() == "foo"
    fs = target.get_fs(rev)
    with fs.open("foo", mode="r") as fobj:
        assert fobj.read().strip() == "foo"


@pytest.mark.skip_git_backend("pygit2")
def test_fetch(tmp_dir: TmpDir, scm: Git, git: Git, tmp_dir_factory: TempDirFactory):
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="init")

    target_dir = tmp_dir_factory.mktemp("git-clone")
    git.clone(str(tmp_dir), (target_dir))
    target = Git(str(target_dir))

    scm.add_commit("bar", message="update")
    rev = scm.get_rev()

    target.fetch()
    assert target.get_ref("refs/remotes/origin/master") == rev


@pytest.mark.skip_git_backend("gitpython")
@pytest.mark.parametrize("untracked_files", ["all", "no", "normal"])
@pytest.mark.parametrize("ignored", [False, True])
def test_status(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
    git_backend,
    tmp_dir_factory: TempDirFactory,
    untracked_files: str,
    ignored: bool,
):
    if untracked_files == "normal" and git_backend == "dulwich":
        pytest.skip(
            "untracked_files=normal is not implemented for dulwich",
        )

    tmp_dir.gen(
        {
            "foo": "foo",
            "bar": "bar",
            ".gitignore": "ignored",
            "ignored": "ignored",
        }
    )
    scm.add_commit(["foo", "bar", ".gitignore"], message="init")

    staged, unstaged, untracked = git.status(ignored, untracked_files)

    assert not staged
    assert not unstaged
    if ignored and untracked_files != "no":
        assert untracked == ["ignored"]
    else:
        assert not untracked

    with (tmp_dir / "foo").open("a") as fobj:
        fobj.write("modified")

    with (tmp_dir / "bar").open("a") as fobj:
        fobj.write("modified")

    tmp_dir.gen({"untracked_dir": {"subfolder": {"subfile": "subfile"}}})
    expected_untracked = []
    if ignored and untracked_files != "no":
        expected_untracked.append("ignored")
    if untracked_files == "all":
        expected_untracked.append("untracked_dir/subfolder/subfile")
    elif untracked_files == "normal":
        expected_untracked.append("untracked_dir/")

    scm.add("foo")
    staged, unstaged, untracked = git.status(ignored, untracked_files)

    assert staged["modify"] == ["foo"]
    assert unstaged == ["bar"]
    assert untracked == expected_untracked


@pytest.mark.skip_git_backend("pygit2")
def test_is_dirty_empty(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
):
    assert not git.is_dirty()


@pytest.mark.skip_git_backend("pygit2")
def test_is_dirty_added(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
):
    tmp_dir.gen("foo", "foo")
    scm.add("foo")

    assert git.is_dirty()
    scm.commit("commit")
    assert not git.is_dirty()


@pytest.mark.skip_git_backend("pygit2")
def test_is_dirty_modified(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
):
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="add foo")

    tmp_dir.gen("foo", "modified")
    assert git.is_dirty()
    scm.add_commit("foo", message="modified")
    assert not git.is_dirty()


@pytest.mark.skip_git_backend("pygit2")
def test_is_dirty_deleted(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
):
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="add foo")

    assert not git.is_dirty()
    os.unlink(tmp_dir / "foo")
    assert git.is_dirty()
    scm.add("foo")
    assert git.is_dirty()
    scm.commit("deleted")
    assert not git.is_dirty()


@pytest.mark.skip_git_backend("pygit2")
def test_is_dirty_untracked(
    tmp_dir: TmpDir,
    scm: Git,
    git: Git,
):
    tmp_dir.gen("untracked", "untracked")
    assert git.is_dirty(untracked_files=True)
    assert not git.is_dirty(untracked_files=False)


@pytest.mark.parametrize(
    "backends", [["gitpython", "dulwich"], ["dulwich", "gitpython"]]
)
def test_backend_func(
    tmp_dir: TmpDir,
    scm: Git,
    backends: list[str],
    mocker: MockerFixture,
):
    from functools import partial

    scm.add = partial(scm._backend_func, "add", backends=backends)
    tmp_dir.gen({"foo": "foo"})
    backend = getattr(scm, backends[0])
    mock = mocker.spy(backend, "add")
    scm.add(["foo"])
    mock.assert_called_once_with(["foo"])


def test_tag(tmp_dir: TmpDir, scm: Git, git: Git):
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="init")
    rev = scm.get_rev()

    git.tag("lightweight")
    assert scm.resolve_rev("lightweight") == rev

    git.tag("annotated", annotated=True, message="message")
    assert scm.resolve_rev("annotated") == rev


def test_get_tag(tmp_dir, scm: Git, git: Git):
    tmp_dir.gen("foo", "foo")
    scm.add_commit("foo", message="init")
    rev = scm.get_rev()

    assert git.get_tag("nonexistent") is None

    scm.tag("lightweight")
    assert git.get_tag("lightweight") == rev

    scm.tag("annotated", annotated=True, message="message")
    tag = git.get_tag("annotated")
    assert isinstance(tag, GitTag)
    assert tag.target == rev
    assert tag.message.strip() == "message"


@pytest.mark.skip_git_backend("gitpython")
def test_config(tmp_dir, scm: Git, git: Git):
    tmp_dir.gen({".git": {"config": "[test]\nfoo = true\n"}})
    tmp_dir.gen(".otherconfig", "[test]\nfoo = false\n")
    config = git.get_config()
    assert config.get(("test",), "foo") == "true"
    assert config.get_bool(("test",), "foo") is True

    config = git.get_config(".otherconfig")
    assert config.get(("test",), "foo") == "false"
    assert config.get_bool(("test",), "foo") is False


@pytest.mark.skip_git_backend("dulwich")
def test_check_attr(tmp_dir, scm: Git, git: Git):
    tmp_dir.gen("foo.txt", "foo")
    tmp_dir.gen(".gitattributes", "*.txt text")
    scm.add_commit([".gitattributes", "foo"], message="init")
    rev = scm.get_rev()
    assert git.check_attr("foo.txt", "text") is True
    assert git.check_attr("foo.txt", "filter") is None

    tmp_dir.gen(".gitattributes", "*.txt -text filter=lfs")
    assert git.check_attr("foo.txt", "text") is False
    assert git.check_attr("foo.txt", "filter") == "lfs"
    assert git.check_attr("foo.txt", "text", source=rev) is True
    assert git.check_attr("foo.txt", "filter", source=rev) is None
